# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import parse_list
from .base import NotifyBase


class NotifyNextcloud(NotifyBase):
    """A wrapper for Nextcloud Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Nextcloud"

    # The services URL
    service_url = "https://nextcloud.com/"

    # Insecure protocol (for those self hosted requests)
    protocol = "ncloud"

    # The default protocol (this is secure for notica)
    secure_protocol = "nclouds"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_nextcloud"

    # Nextcloud title length
    title_maxlen = 255

    # Defines the maximum allowable characters per message.
    body_maxlen = 4000

    # Define object templates
    templates = (
        "{schema}://{host}/{targets}",
        "{schema}://{host}:{port}/{targets}",
        "{schema}://{user}:{password}@{host}/{targets}",
        "{schema}://{user}:{password}@{host}:{port}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
            "user": {
                "name": _("Username"),
                "type": "string",
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
            },
            "target_user": {
                "name": _("Target User"),
                "type": "string",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            # Nextcloud uses different API end points depending on the version
            # being used however the (API) payload remains the same.  Allow
            # users to specify the version they are using:
            "version": {
                "name": _("Version"),
                "type": "int",
                "min": 1,
                "default": 21,
            },
            "url_prefix": {
                "name": _("URL Prefix"),
                "type": "string",
            },
            "to": {
                "alias_of": "targets",
            },
        },
    )

    # Define any kwargs we're using
    template_kwargs = {
        "headers": {
            "name": _("HTTP Header"),
            "prefix": "+",
        },
    }

    def __init__(
        self,
        targets=None,
        version=None,
        headers=None,
        url_prefix=None,
        **kwargs,
    ):
        """Initialize Nextcloud Object."""
        super().__init__(**kwargs)

        # Store our targets
        self.targets = parse_list(targets)

        self.version = self.template_args["version"]["default"]
        if version is not None:
            try:
                self.version = int(version)
                if self.version < self.template_args["version"]["min"]:
                    # Let upper exception handle this
                    raise ValueError()

            except (ValueError, TypeError):
                msg = (
                    f"At invalid Nextcloud version ({version}) was specified."
                )
                self.logger.warning(msg)
                raise TypeError(msg) from None

        # Support URL Prefix
        self.url_prefix = "" if not url_prefix else url_prefix.strip("/")

        self.headers = {}
        if headers:
            # Store our extra headers
            self.headers.update(headers)

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Nextcloud Notification."""

        if len(self.targets) == 0:
            # There were no services to notify
            self.logger.warning("There were no Nextcloud targets to notify.")
            return False

        # Prepare our Header
        headers = {
            "User-Agent": self.app_id,
            "OCS-APIREQUEST": "true",
        }

        # Apply any/all header over-rides defined
        headers.update(self.headers)

        # error tracking (used for function return)
        has_error = False

        # Create a copy of the targets list
        targets = list(self.targets)
        while len(targets):
            target = targets.pop(0)

            # Prepare our Payload
            payload = {
                "shortMessage": title if title else self.app_desc,
            }
            if body:
                # Only store the longMessage if a body was defined; nextcloud
                # doesn't take kindly to empty longMessage entries.
                payload["longMessage"] = body

            auth = None
            if self.user:
                auth = (self.user, self.password)

            # Nextcloud URL based on version used
            notify_url = (
                "{schema}://{host}/{url_prefix}/ocs/v2.php/"
                "apps/admin_notifications/"
                "api/v1/notifications/{target}"
                if self.version < 21
                else (
                    "{schema}://{host}/{url_prefix}/ocs/v2.php/"
                    "apps/notifications/"
                    "api/v2/admin_notifications/{target}"
                )
            )

            notify_url = notify_url.format(
                schema="https" if self.secure else "http",
                host=(
                    self.host
                    if not isinstance(self.port, int)
                    else f"{self.host}:{self.port}"
                ),
                url_prefix=self.url_prefix,
                target=target,
            )

            self.logger.debug(
                "Nextcloud v%d POST URL: %s (cert_verify=%r)",
                self.version,
                notify_url,
                self.verify_certificate,
            )
            self.logger.debug(
                "Nextcloud v%d Payload: %s", self.version, str(payload)
            )

            # Always call throttle before any remote server i/o is made
            self.throttle()

            try:
                r = requests.post(
                    notify_url,
                    data=payload,
                    headers=headers,
                    auth=auth,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )
                if r.status_code != requests.codes.ok:
                    # We had a problem
                    status_str = NotifyNextcloud.http_response_code_lookup(
                        r.status_code
                    )

                    self.logger.warning(
                        "Failed to send Nextcloud v{} notification:"
                        "{}{}error={}.".format(
                            self.version,
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")
                    # track our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(
                        "Sent Nextcloud %d notification.", self.version
                    )

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending Nextcloud v%d"
                    "notification.",
                    self.version,
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # track our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.user,
            self.password,
            self.host,
            self.port,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Create URL parameters from our headers
        params = {f"+{k}": v for k, v in self.headers.items()}

        # Set our version
        params["version"] = str(self.version)

        if self.url_prefix:
            params["url_prefix"] = self.url_prefix

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Determine Authentication
        auth = ""
        if self.user and self.password:
            auth = "{user}:{password}@".format(
                user=NotifyNextcloud.quote(self.user, safe=""),
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=""
                ),
            )
        elif self.user:
            auth = "{user}@".format(
                user=NotifyNextcloud.quote(self.user, safe=""),
            )

        default_port = 443 if self.secure else 80
        return "{schema}://{auth}{hostname}{port}/{targets}?{params}".format(
            schema=self.secure_protocol if self.secure else self.protocol,
            auth=auth,
            # never encode hostname since we're expecting it to be a
            # valid one
            hostname=self.host,
            port=(
                ""
                if self.port is None or self.port == default_port
                else f":{self.port}"
            ),
            targets="/".join([NotifyNextcloud.quote(x) for x in self.targets]),
            params=NotifyNextcloud.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        targets = len(self.targets)
        return targets if targets else 1

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Fetch our targets
        results["targets"] = NotifyNextcloud.split_path(results["fullpath"])

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyNextcloud.parse_list(
                results["qsd"]["to"]
            )

        # Allow users to over-ride the Nextcloud version being used
        if "version" in results["qsd"] and len(results["qsd"]["version"]):
            results["version"] = NotifyNextcloud.unquote(
                results["qsd"]["version"]
            )

        # Support URL Prefixes
        if "url_prefix" in results["qsd"] and len(
            results["qsd"]["url_prefix"]
        ):
            results["url_prefix"] = NotifyNextcloud.unquote(
                results["qsd"]["url_prefix"]
            )

        # Add our headers that the user can potentially over-ride if they wish
        # to to our returned result set and tidy entries by unquoting them
        results["headers"] = {
            NotifyNextcloud.unquote(x): NotifyNextcloud.unquote(y)
            for x, y in results["qsd+"].items()
        }

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from os.path import abspath, dirname, isfile, join
import re
from typing import Any, Optional, Union
from uuid import uuid4

from .common import (
    NotifyFormat,
    NotifyImageSize,
    NotifyType,
    PersistentStoreMode,
)
from .manager_plugins import NotificationManager

# Grant access to our Notification Manager Singleton
N_MGR = NotificationManager()


class AppriseAsset:
    """Provides a supplimentary class that can be used to provide extra
    information and details that can be used by Apprise such as providing an
    alternate location to where images/icons can be found and the URL masks.

    Any variable that starts with an underscore (_) can only be initialized by
    this class manually and will/can not be parsed from a configuration file.
    """

    # Application Identifier
    app_id = "Apprise"

    # Application Description
    app_desc = "Apprise Notifications"

    # Provider URL
    app_url = "https://github.com/caronc/apprise"

    # A Simple Mapping of Colors; For every NOTIFY_TYPE identified,
    # there should be a mapping to it's color here:
    html_notify_map = {
        NotifyType.INFO: "#3AA3E3",
        NotifyType.SUCCESS: "#3AA337",
        NotifyType.FAILURE: "#A32037",
        NotifyType.WARNING: "#CACF29",
    }

    # The default color to return if a mapping isn't found in our table above
    default_html_color = "#888888"

    # Ascii Notification
    ascii_notify_map = {
        NotifyType.INFO: "[i]",
        NotifyType.SUCCESS: "[+]",
        NotifyType.FAILURE: "[!]",
        NotifyType.WARNING: "[~]",
    }

    # The default ascii to return if a mapping isn't found in our table above
    default_ascii_chars = "[?]"

    # The default image extension to use
    default_extension = ".png"

    # The default image size if one isn't specified
    default_image_size = NotifyImageSize.XY_256

    # The default theme
    theme = "default"

    # Image URL Mask
    image_url_mask = (
        "https://github.com/caronc/apprise/raw/master/apprise/assets/"
        "themes/{THEME}/apprise-{TYPE}-{XY}{EXTENSION}"
    )

    # Application Logo
    image_url_logo = (
        "https://github.com/caronc/apprise/raw/master/apprise/assets/"
        "themes/{THEME}/apprise-logo.png"
    )

    # Image Path Mask
    image_path_mask = abspath(
        join(
            dirname(__file__),
            "assets",
            "themes",
            "{THEME}",
            "apprise-{TYPE}-{XY}{EXTENSION}",
        )
    )

    # This value can also be set on calls to Apprise.notify(). This allows
    # you to let Apprise upfront the type of data being passed in.  This
    # must be of type NotifyFormat. Possible values could be:
    # - NotifyFormat.TEXT
    # - NotifyFormat.MARKDOWN
    # - NotifyFormat.HTML
    # - None
    #
    # If no format is specified (hence None), then no special pre-formatting
    # actions will take place during a notification. This has been and always
    # will be the default.
    body_format = None

    # Always attempt to send notifications asynchronous (as the same time
    # if possible)
    # This is a Python 3 supported option only. If set to False, then
    # notifications are sent sequentially (one after another)
    async_mode = True

    # Support :smile:, and other alike keywords swapping them for their
    # unicode value. A value of None leaves the interpretation up to the
    # end user to control (allowing them to specify emojis=yes on the
    # URL)
    interpret_emojis = None

    # Whether or not to interpret escapes found within the input text prior
    # to passing it upstream. Such as converting \t to an actual tab and \n
    # to a new line.
    interpret_escapes = False

    # Defines the encoding of the content passed into Apprise
    encoding = "utf-8"

    # Automatically generate our Pretty Good Privacy (PGP) keys if one isn't
    # present and our environment configuration allows for it.
    # For example, a case where the environment wouldn't allow for it would be
    # if Persistent Storage was set to `memory`
    pgp_autogen = True

    # Automatically generate our Privacy Enhanced Mail (PEM) keys if one isn't
    # present and our environment configuration allows for it.
    # For example, a case where the environment wouldn't allow for it would be
    # if Persistent Storage was set to `memory`
    pem_autogen = True

    # For more detail see CWE-312 @
    #    https://cwe.mitre.org/data/definitions/312.html
    #
    # By enabling this, the logging output has additional overhead applied to
    # it preventing secure password and secret information from being
    # displayed in the logging. Since there is overhead involved in performing
    # this cleanup; system owners who run in a very isolated environment may
    # choose to disable this for a slight performance bump. It is recommended
    # that you leave this option as is otherwise.
    secure_logging = True

    # Optionally specify one or more path to attempt to scan for Python modules
    # By default, no paths are scanned.
    __plugin_paths = []

    # Optionally set the location of the persistent storage
    # By default there is no path and thus persistent storage is not used
    __storage_path = None

    # Optionally define the default salt to apply to all persistent storage
    # namespace generation (unless over-ridden)
    __storage_salt = b""

    # Optionally define the namespace length of the directories created by
    # the storage. If this is set to zero, then the length is pre-determined
    # by the generator (sha1, md5, sha256, etc)
    __storage_idlen = 8

    # Set storage to auto
    __storage_mode = PersistentStoreMode.AUTO

    # All internal/system flags are prefixed with an underscore (_)
    # These can only be initialized using Python libraries and are not picked
    # up from (yaml) configuration files (if set)

    # An internal counter that is used by AppriseAPI
    # (https://github.com/caronc/apprise-api). The idea is to allow one
    # instance of AppriseAPI to call another, but to track how many times
    # this occurs. It's intent is to prevent a loop where an AppriseAPI
    # Server calls itself (or loops indefinitely)
    _recursion = 0

    # A unique identifer we can use to associate our calling source
    _uid = str(uuid4())

    def __init__(
        self,
        plugin_paths: Optional[list[str]] = None,
        storage_path: Optional[str] = None,
        storage_mode: Optional[Union[str, PersistentStoreMode]] = None,
        storage_salt: Optional[Union[str, bytes]] = None,
        storage_idlen: Optional[int] = None,
        **kwargs: Any
    ) -> None:
        """Asset Initialization."""
        # Assign default arguments if specified
        for key, value in kwargs.items():
            if not hasattr(AppriseAsset, key):
                raise AttributeError(
                    f"AppriseAsset init(): An invalid key {key} was specified."
                )

            setattr(self, key, value)

        if plugin_paths:
            # Load any decorated modules if defined
            self.__plugin_paths = plugin_paths
            N_MGR.module_detection(plugin_paths)

        if storage_path:
            # Define our persistent storage path
            self.__storage_path = storage_path

        if storage_mode:
            # Define how our persistent storage behaves
            try:
                self.__storage_mode = (
                    storage_mode if isinstance(storage_mode, NotifyFormat)
                    else PersistentStoreMode(storage_mode.lower())
                )

            except (AttributeError, ValueError, TypeError):
                err = (
                    f"An invalid persistent store mode ({storage_mode}) was "
                    "specified.")
                raise AttributeError(err) from None

        if isinstance(storage_idlen, int):
            # Define the number of characters utilized from our namespace lengh
            if storage_idlen < 0:
                # Unsupported type
                raise ValueError(
                    "AppriseAsset storage_idlen(): Value must "
                    "be an integer and > 0"
                )

            # Store value
            self.__storage_idlen = storage_idlen

        if storage_salt is not None:
            # Define the number of characters utilized from our namespace lengh

            if isinstance(storage_salt, bytes):
                self.__storage_salt = storage_salt

            elif isinstance(storage_salt, str):
                try:
                    self.__storage_salt = storage_salt.encode(self.encoding)

                except UnicodeEncodeError:
                    # Bad data; don't pass it along
                    raise ValueError(
                        "AppriseAsset namespace_salt(): "
                        "Value provided could not be encoded"
                    ) from None

            else:  # Unsupported
                raise ValueError(
                    "AppriseAsset namespace_salt(): Value provided must be "
                    "string or bytes object"
                )

    def color(
        self,
        notify_type: NotifyType,
        color_type: Optional[type] = None,
    ) -> Union[str, int, tuple[int, int, int]]:
        """Returns an HTML mapped color based on passed in notify type.

        if color_type is:
           None    then a standard hex string is returned as
                   a string format ('#000000').

           int     then the integer representation is returned
           tuple   then the the red, green, blue is returned in a tuple
        """

        # Attempt to get the type, otherwise return a default grey
        # if we couldn't look up the entry
        color = self.html_notify_map.get(
            notify_type, self.default_html_color)
        if color_type is None:
            # This is the default return type
            return color

        elif color_type is int:
            # Convert the color to integer
            return AppriseAsset.hex_to_int(color)

        # The only other type is tuple
        elif color_type is tuple:
            return AppriseAsset.hex_to_rgb(color)

        # Unsupported type
        raise ValueError(
            "AppriseAsset html_color(): An invalid color_type was specified."
        )

    def ascii(self, notify_type: NotifyType) -> str:
        """Returns an ascii representation based on passed in notify type."""
        # look our response up
        return self.ascii_notify_map.get(
            notify_type, self.default_ascii_chars)

    def image_url(
        self,
        notify_type: NotifyType,
        image_size: Optional[NotifyImageSize] = None,
        logo: bool = False,
        extension: Optional[str] = None,
    ) -> Optional[str]:
        """Apply our mask to our image URL.

        if logo is set to True, then the logo_url is used instead
        """

        url_mask = self.image_url_logo if logo else self.image_url_mask
        if not url_mask:
            # No image to return
            return None

        if extension is None:
            extension = self.default_extension

        if image_size is None:
            image_size = self.default_image_size

        re_map = {
            "{THEME}": self.theme if self.theme else "",
            "{TYPE}": notify_type.value,
            "{XY}": image_size.value,
            "{EXTENSION}": extension,
        }

        # Iterate over above list and store content accordingly
        re_table = re.compile(
            r"(" + "|".join(re_map.keys()) + r")",
            re.IGNORECASE,
        )

        return re_table.sub(lambda x: re_map[x.group()], url_mask)

    def image_path(
        self,
        notify_type: NotifyType,
        image_size: NotifyImageSize,
        must_exist: bool = True,
        extension: Optional[str] = None,
    ) -> Optional[str]:
        """Apply our mask to our image file path."""

        if not self.image_path_mask:
            # No image to return
            return None

        if extension is None:
            extension = self.default_extension

        re_map = {
            "{THEME}": self.theme if self.theme else "",
            "{TYPE}": notify_type.value,
            "{XY}": image_size.value,
            "{EXTENSION}": extension,
        }

        # Iterate over above list and store content accordingly
        re_table = re.compile(
            r"(" + "|".join(re_map.keys()) + r")",
            re.IGNORECASE,
        )

        # Acquire our path
        path = re_table.sub(lambda x: re_map[x.group()], self.image_path_mask)
        if must_exist and not isfile(path):
            return None

        # Return what we parsed
        return path

    def image_raw(
        self,
        notify_type: NotifyType,
        image_size: NotifyImageSize,
        extension: Optional[str] = None,
    ) -> Optional[bytes]:
        """Returns the raw image if it can (otherwise the function returns
        None)"""

        path = self.image_path(
            notify_type=notify_type,
            image_size=image_size,
            extension=extension,
        )
        if path:
            try:
                with open(path, "rb") as fd:
                    return fd.read()

            except OSError:
                # We can't access the file
                return None

        return None

    def details(self) -> dict[str, str]:
        """Returns the details associated with the AppriseAsset object."""
        return {
            "app_id": self.app_id,
            "app_desc": self.app_desc,
            "default_extension": self.default_extension,
            "theme": self.theme,
            "image_path_mask": self.image_path_mask,
            "image_url_mask": self.image_url_mask,
            "image_url_logo": self.image_url_logo,
        }

    @staticmethod
    def hex_to_rgb(value: str) -> tuple[int, int, int]:
        """Takes a hex string (such as #00ff00) and returns a tuple in the form
        of (red, green, blue)

        eg: #00ff00 becomes : (0, 65535, 0)
        """
        value = value.lstrip("#")
        lv = len(value)
        return tuple(
            int(value[i : i + lv // 3], 16) for i in range(0, lv, lv // 3)
        )

    @staticmethod
    def hex_to_int(value: str) -> int:
        """Takes a hex string (such as #00ff00) and returns its integer
        equivalent.

        eg: #00000f becomes : 15
        """
        return int(value.lstrip("#"), 16)

    @property
    def plugin_paths(self) -> list[str]:
        """Return the plugin paths defined."""
        return self.__plugin_paths

    @property
    def storage_path(self) -> Optional[str]:
        """Return the persistent storage path defined."""
        return self.__storage_path

    @property
    def storage_mode(self) -> PersistentStoreMode:
        """Return the persistent storage mode defined."""

        return self.__storage_mode

    @property
    def storage_salt(self) -> bytes:
        """Return the provided namespace salt; this is always of type bytes."""
        return self.__storage_salt

    @property
    def storage_idlen(self) -> int:
        """Return the persistent storage id length."""

        return self.__storage_idlen



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# For this to work correctly you need to create a webhook. To do this just
# click on the little gear icon next to the channel you're part of. From
# here you'll be able to access the Webhooks menu and create a new one.
#
#  When you've completed, you'll get a URL that looks a little like this:
#  https://discord.com/api/webhooks/417429632418316298/\
#         JHZ7lQml277CDHmQKMHI8qBe7bk2ZwO5UKjCiOAF7711o33MyqU344Qpgv7YTpadV_js
#
#  Simplified, it looks like this:
#     https://discord.com/api/webhooks/WEBHOOK_ID/WEBHOOK_TOKEN
#
#  This plugin will simply work using the url of:
#     discord://WEBHOOK_ID/WEBHOOK_TOKEN
#
# API Documentation on Webhooks:
#    - https://discord.com/developers/docs/resources/webhook
#
from datetime import datetime, timedelta, timezone
from json import dumps
import re

import requests

from ..attachment.base import AttachBase
from ..common import NotifyFormat, NotifyImageSize, NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import parse_bool, validate_regex
from .base import NotifyBase

# Used to detect user/role IDs
USER_ROLE_DETECTION_RE = re.compile(
    r"\s*(?:<@(?P<role>&?)(?P<id>[0-9]+)>|@(?P<value>[a-z0-9]+))", re.I
)


class NotifyDiscord(NotifyBase):
    """A wrapper to Discord Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Discord"

    # The services URL
    service_url = "https://discord.com/"

    # The default secure protocol
    secure_protocol = "discord"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_discord"

    # Discord Webhook
    notify_url = "https://discord.com/api/webhooks"

    # Support attachments
    attachment_support = True

    # Allows the user to specify the NotifyImageSize object
    image_size = NotifyImageSize.XY_256

    # Discord is kind enough to return how many more requests we're allowed to
    # continue to make within it's header response as:
    # X-RateLimit-Reset: The epoc time (in seconds) we can expect our
    #                    rate-limit to be reset.
    # X-RateLimit-Remaining: an integer identifying how many requests we're
    #                        still allow to make.
    request_rate_per_sec = 0

    # Taken right from google.auth.helpers:
    clock_skew = timedelta(seconds=10)

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 2000

    # The 2000 characters above defined by the body_maxlen include that of the
    # title.  Setting this to True ensures overflow options behave properly
    overflow_amalgamate_title = True

    # Discord has a limit of the number of fields you can include in an
    # embeds message. This value allows the discord message to safely
    # break into multiple messages to handle these cases.
    discord_max_fields = 10

    # Define object templates
    templates = (
        "{schema}://{webhook_id}/{webhook_token}",
        "{schema}://{botname}@{webhook_id}/{webhook_token}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "botname": {
                "name": _("Bot Name"),
                "type": "string",
                "map_to": "user",
            },
            "webhook_id": {
                "name": _("Webhook ID"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "webhook_token": {
                "name": _("Webhook Token"),
                "type": "string",
                "private": True,
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "tts": {
                "name": _("Text To Speech"),
                "type": "bool",
                "default": False,
            },
            "avatar": {
                "name": _("Avatar Image"),
                "type": "bool",
                "default": True,
            },
            "avatar_url": {
                "name": _("Avatar URL"),
                "type": "string",
            },
            "href": {
                "name": _("URL"),
                "type": "string",
            },
            "url": {
                "alias_of": "href",
            },
            # Send a message to the specified thread within a webhook's
            # channel. The thread will automatically be unarchived.
            "thread": {
                "name": _("Thread ID"),
                "type": "string",
            },
            "footer": {
                "name": _("Display Footer"),
                "type": "bool",
                "default": False,
            },
            "footer_logo": {
                "name": _("Footer Logo"),
                "type": "bool",
                "default": True,
            },
            "fields": {
                "name": _("Use Fields"),
                "type": "bool",
                "default": True,
            },
            "image": {
                "name": _("Include Image"),
                "type": "bool",
                "default": False,
                "map_to": "include_image",
            },
        },
    )

    def __init__(
        self,
        webhook_id,
        webhook_token,
        tts=False,
        avatar=True,
        footer=False,
        footer_logo=True,
        include_image=False,
        fields=True,
        avatar_url=None,
        href=None,
        thread=None,
        **kwargs,
    ):
        """Initialize Discord Object."""
        super().__init__(**kwargs)

        # Webhook ID (associated with project)
        self.webhook_id = validate_regex(webhook_id)
        if not self.webhook_id:
            msg = (
                f"An invalid Discord Webhook ID ({webhook_id}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Webhook Token (associated with project)
        self.webhook_token = validate_regex(webhook_token)
        if not self.webhook_token:
            msg = (
                "An invalid Discord Webhook Token "
                f"({webhook_token}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Text To Speech
        self.tts = tts

        # Over-ride Avatar Icon
        self.avatar = avatar

        # Place a footer
        self.footer = footer

        # include a footer_logo in footer
        self.footer_logo = footer_logo

        # Place a thumbnail image inline with the message body
        self.include_image = include_image

        # Use Fields
        self.fields = fields

        # Specified Thread ID
        self.thread_id = thread

        # Avatar URL
        # This allows a user to provide an over-ride to the otherwise
        # dynamically generated avatar url images
        self.avatar_url = avatar_url

        # A URL to have the title link to
        self.href = href

        # For Tracking Purposes
        self.ratelimit_reset = datetime.now(timezone.utc).replace(tzinfo=None)

        # Default to 1.0
        self.ratelimit_remaining = 1.0

        return

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Perform Discord Notification."""

        payload = {
            # Text-To-Speech
            "tts": self.tts,
            # If Text-To-Speech is set to True, then we do not want to wait
            # for the whole message before continuing. Otherwise, we wait
            "wait": self.tts is False,
        }

        # Acquire image_url
        image_url = self.image_url(notify_type)

        if self.avatar and (image_url or self.avatar_url):
            payload["avatar_url"] = (
                self.avatar_url if self.avatar_url else image_url
            )

        if self.user:
            # Optionally override the default username of the webhook
            payload["username"] = self.user

        # Associate our thread_id with our message
        params = {"thread_id": self.thread_id} if self.thread_id else None

        if body:
            # our fields variable
            fields = []

            if self.notify_format == NotifyFormat.MARKDOWN:
                # Use embeds for payload
                payload["embeds"] = [{
                    "author": {
                        "name": self.app_id,
                        "url": self.app_url,
                    },
                    "title": title,
                    "description": body,
                    # Our color associated with our notification
                    "color": self.color(notify_type, int),
                }]

                if self.href:
                    payload["embeds"][0]["url"] = self.href

                if self.footer:
                    # Acquire logo URL
                    logo_url = self.image_url(notify_type, logo=True)

                    # Set Footer text to our app description
                    payload["embeds"][0]["footer"] = {
                        "text": self.app_desc,
                    }

                    if self.footer_logo and logo_url:
                        payload["embeds"][0]["footer"]["icon_url"] = logo_url

                if self.include_image and image_url:
                    payload["embeds"][0]["thumbnail"] = {
                        "url": image_url,
                        "height": 256,
                        "width": 256,
                    }

                if self.fields:
                    # Break titles out so that we can sort them in embeds
                    description, fields = self.extract_markdown_sections(body)

                    # Swap first entry for description
                    payload["embeds"][0]["description"] = description
                    if fields:
                        # Apply our additional parsing for a better
                        # presentation
                        payload["embeds"][0]["fields"] = fields[
                            : self.discord_max_fields
                        ]

                        # Remove entry from head of fields
                        fields = fields[self.discord_max_fields:]

            else:
                # not markdown
                payload["content"] = (
                    body if not title else f"{title}\r\n{body}"
                )

            # parse for user id's <@123> and role IDs <@&456>
            results = USER_ROLE_DETECTION_RE.findall(body)
            if results:
                payload["allow_mentions"] = {
                    "parse": [],
                    "users": [],
                    "roles": [],
                }

                _content = []
                for is_role, no, value in results:
                    if value:
                        payload["allow_mentions"]["parse"].append(value)
                        _content.append(f"@{value}")

                    elif is_role:
                        payload["allow_mentions"]["roles"].append(no)
                        _content.append(f"<@&{no}>")

                    else:  # is_user
                        payload["allow_mentions"]["users"].append(no)
                        _content.append(f"<@{no}>")

                if self.notify_format == NotifyFormat.MARKDOWN:
                    # Add pingable elements to content field
                    payload["content"] = "👉 " + " ".join(_content)

            if not self._send(payload, params=params):
                # We failed to post our message
                return False

            # Process any remaining fields IF set
            if fields:
                payload["embeds"][0]["description"] = ""
                for i in range(0, len(fields), self.discord_max_fields):
                    payload["embeds"][0]["fields"] = fields[
                        i : i + self.discord_max_fields
                    ]
                    if not self._send(payload):
                        # We failed to post our message
                        return False

        if attach and self.attachment_support:
            # Update our payload; the idea is to preserve it's other detected
            # and assigned values for re-use here too
            payload.update({
                # Text-To-Speech
                "tts": False,
                # Wait until the upload has posted itself before continuing
                "wait": True,
            })

            #
            # Remove our text/title based content for attachment use
            #
            payload.pop("embeds", None)

            payload.pop("content", None)

            payload.pop("allow_mentions", None)

            #
            # Send our attachments
            #
            for attachment in attach:
                self.logger.info(
                    f"Posting Discord Attachment {attachment.name}"
                )
                if not self._send(payload, params=params, attach=attachment):
                    # We failed to post our message
                    return False

        # Otherwise return
        return True

    def _send(self, payload, attach=None, params=None, rate_limit=1, **kwargs):
        """Wrapper to the requests (post) object."""

        # Our headers
        headers = {
            "User-Agent": self.app_id,
        }

        # Construct Notify URL
        notify_url = (
            f"{self.notify_url}/{self.webhook_id}/{self.webhook_token}"
        )

        self.logger.debug(
            "Discord POST URL:"
            f" {notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Discord Payload: {payload!s}")

        # By default set wait to None
        wait = None

        if self.ratelimit_remaining <= 0.0:
            # Determine how long we should wait for or if we should wait at
            # all. This isn't fool-proof because we can't be sure the client
            # time (calling this script) is completely synced up with the
            # Discord server.  One would hope we're on NTP and our clocks are
            # the same allowing this to role smoothly:

            now = datetime.now(timezone.utc).replace(tzinfo=None)
            if now < self.ratelimit_reset:
                # We need to throttle for the difference in seconds
                wait = abs(
                    (
                        self.ratelimit_reset - now + self.clock_skew
                    ).total_seconds()
                )

        # Always call throttle before any remote server i/o is made;
        self.throttle(wait=wait)

        # Perform some simple error checking
        if isinstance(attach, AttachBase):
            if not attach:
                # We could not access the attachment
                self.logger.error(
                    f"Could not access attachment {attach.url(privacy=True)}."
                )
                return False

            self.logger.debug(
                f"Posting Discord attachment {attach.url(privacy=True)}"
            )

        # Our attachment path (if specified)
        files = None
        try:

            # Open our attachment path if required:
            if attach:
                files = {
                    "file": (
                        attach.name,
                        # file handle is safely closed in `finally`; inline
                        # open is intentional
                        open(attach.path, "rb"),  # noqa: SIM115
                    )}

            else:
                headers["Content-Type"] = "application/json; charset=utf-8"

            r = requests.post(
                notify_url,
                params=params,
                data=payload if files else dumps(payload),
                headers=headers,
                files=files,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            # Handle rate limiting (if specified)
            try:
                # Store our rate limiting (if provided)
                self.ratelimit_remaining = float(
                    r.headers.get("X-RateLimit-Remaining")
                )
                self.ratelimit_reset = datetime.fromtimestamp(
                    int(r.headers.get("X-RateLimit-Reset")), timezone.utc
                ).replace(tzinfo=None)

            except (TypeError, ValueError):
                # This is returned if we could not retrieve this
                # information gracefully accept this state and move on
                pass

            if r.status_code not in (
                requests.codes.ok,
                requests.codes.no_content,
            ):

                # We had a problem
                status_str = NotifyBase.http_response_code_lookup(
                    r.status_code
                )

                if (
                    r.status_code == requests.codes.too_many_requests
                    and rate_limit > 0
                ):

                    # handle rate limiting
                    self.logger.warning(
                        "Discord rate limiting in effect; "
                        "blocking for %.2f second(s)",
                        self.ratelimit_remaining,
                    )

                    # Try one more time before failing
                    return self._send(
                        payload=payload,
                        attach=attach,
                        params=params,
                        rate_limit=rate_limit - 1,
                        **kwargs,
                    )

                self.logger.warning(
                    "Failed to send {}to Discord notification: "
                    "{}{}error={}.".format(
                        attach.name if attach else "",
                        status_str,
                        ", " if status_str else "",
                        r.status_code,
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info(
                    "Sent Discord {}.".format(
                        "attachment" if attach else "notification"
                    )
                )

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred posting {}to Discord.".format(
                    attach.name if attach else ""
                )
            )
            self.logger.debug(f"Socket Exception: {e!s}")
            return False

        except OSError as e:
            self.logger.warning(
                "An I/O error occurred while reading {}.".format(
                    attach.name if attach else "attachment"
                )
            )
            self.logger.debug(f"I/O Exception: {e!s}")
            return False

        finally:
            # Close our file (if it's open) stored in the second element
            # of our files tuple (index 1)
            if files:
                files["file"][1].close()

        return True

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "tts": "yes" if self.tts else "no",
            "avatar": "yes" if self.avatar else "no",
            "footer": "yes" if self.footer else "no",
            "footer_logo": "yes" if self.footer_logo else "no",
            "image": "yes" if self.include_image else "no",
            "fields": "yes" if self.fields else "no",
        }

        if self.avatar_url:
            params["avatar_url"] = self.avatar_url

        if self.href:
            params["href"] = self.href

        if self.thread_id:
            params["thread"] = self.thread_id

        # Ensure our botname is set
        botname = f"{self.user}@" if self.user else ""

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return (
            "{schema}://{botname}{webhook_id}/{webhook_token}/?{params}"
            .format(
                schema=self.secure_protocol,
                botname=botname,
                webhook_id=self.pprint(self.webhook_id, privacy, safe=""),
                webhook_token=self.pprint(
                    self.webhook_token, privacy, safe=""
                ),
                params=NotifyDiscord.urlencode(params),
            )
        )

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.webhook_id, self.webhook_token)

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object.

        Syntax:
          discord://webhook_id/webhook_token
        """
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Store our webhook ID
        webhook_id = NotifyDiscord.unquote(results["host"])

        # Now fetch our tokens
        try:
            webhook_token = NotifyDiscord.split_path(results["fullpath"])[0]

        except IndexError:
            # Force some bad values that will get caught
            # in parsing later
            webhook_token = None

        results["webhook_id"] = webhook_id
        results["webhook_token"] = webhook_token

        # Text To Speech
        results["tts"] = parse_bool(results["qsd"].get("tts", False))

        # Use sections
        # effectively detect multiple fields and break them off
        # into sections
        results["fields"] = parse_bool(results["qsd"].get("fields", True))

        # Use Footer
        results["footer"] = parse_bool(results["qsd"].get("footer", False))

        # Use Footer Logo
        results["footer_logo"] = parse_bool(
            results["qsd"].get("footer_logo", True)
        )

        # Update Avatar Icon
        results["avatar"] = parse_bool(results["qsd"].get("avatar", True))

        # Boolean to include an image or not
        results["include_image"] = parse_bool(
            results["qsd"].get(
                "image", NotifyDiscord.template_args["image"]["default"]
            )
        )

        if "botname" in results["qsd"]:
            # Alias to User
            results["user"] = NotifyDiscord.unquote(results["qsd"]["botname"])

        # Extract avatar url if it was specified
        if "avatar_url" in results["qsd"]:
            results["avatar_url"] = NotifyDiscord.unquote(
                results["qsd"]["avatar_url"]
            )

        # Extract url if it was specified
        if "href" in results["qsd"]:
            results["href"] = NotifyDiscord.unquote(results["qsd"]["href"])

        elif "url" in results["qsd"]:
            results["href"] = NotifyDiscord.unquote(results["qsd"]["url"])
            # Markdown is implied
            results["format"] = NotifyFormat.MARKDOWN

        # Extract thread id if it was specified
        if "thread" in results["qsd"]:
            results["thread"] = NotifyDiscord.unquote(results["qsd"]["thread"])
            # Markdown is implied
            results["format"] = NotifyFormat.MARKDOWN

        return results

    @staticmethod
    def parse_native_url(url):
        """
        Support https://discord.com/api/webhooks/WEBHOOK_ID/WEBHOOK_TOKEN
        Support Legacy URL as well:
            https://discordapp.com/api/webhooks/WEBHOOK_ID/WEBHOOK_TOKEN
        """

        result = re.match(
            r"^https?://discord(app)?\.com/api/webhooks/"
            r"(?P<webhook_id>[0-9]+)/"
            r"(?P<webhook_token>[A-Z0-9_-]+)/?"
            r"(?P<params>\?.+)?$",
            url,
            re.I,
        )

        if result:
            return NotifyDiscord.parse_url(
                "{schema}://{webhook_id}/{webhook_token}/{params}".format(
                    schema=NotifyDiscord.secure_protocol,
                    webhook_id=result.group("webhook_id"),
                    webhook_token=result.group("webhook_token"),
                    params=(
                        ""
                        if not result.group("params")
                        else result.group("params")
                    ),
                )
            )

        return None

    @staticmethod
    def extract_markdown_sections(markdown):
        """Takes a string in a markdown type format and extracts the headers
        and their corresponding sections into individual fields that get passed
        as an embed entry to Discord."""
        # Search for any header information found without it's own section
        # identifier
        match = re.match(
            r"^\s*(?P<desc>[^\s#]+.*?)(?=\s*$|[\r\n]+\s*#)",
            markdown,
            flags=re.S,
        )

        description = match.group("desc").strip() if match else ""
        if description:
            # Strip description from our string since it has been handled
            # now.
            markdown = re.sub(re.escape(description), "", markdown, count=1)

        regex = re.compile(
            r"\s*#[# \t\v]*(?P<name>[^\n]+)(\n|\s*$)"
            r"\s*((?P<value>[^#].+?)(?=\s*$|[\r\n]+\s*#))?",
            flags=re.S,
        )

        common = regex.finditer(markdown)
        fields = []
        for el in common:
            d = el.groupdict()

            fields.append({
                "name": d.get("name", "").strip("#`* \r\n\t\v"),
                "value": "```{}\n{}```".format(
                    "md" if d.get("value") else "",
                    (d.get("value").strip() + "\n" if d.get("value") else ""),
                ),
            })

        return description, fields
